#lang scheme/gui

(require "board.ss"
         "play.ss")

(define SIZE 21)

;; Offscreen bitmap (to avoid flicker):
(define dc (make-object bitmap-dc% (make-object bitmap% 1 1)))
(define (reset-offscreen canvas-dc)
  ;; Make sure ofscreen bitmap matches canvas size
  (let-values ([(w h) (send canvas-dc get-size)]
               [(bw bh) (send dc get-size)])
    (unless (and (= w bw) (= h bh))
      (send dc set-bitmap (make-object bitmap% 
                            (inexact->exact w) 
                            (inexact->exact h))))))

;; Board-drawing function:
(define (draw-board canvas-dc)
  (define (center-string color s)
    (send dc set-text-foreground (make-object color% color))
    (let-values ([(w h d a) (send dc get-text-extent s)])
      (send dc draw-text s (/ (- SIZE w) 2) (/ (- SIZE h d a) 2))))
  (reset-offscreen canvas-dc)
  (send dc clear)
  (send dc set-pen "white" 0 'transparent)
  ;; Draw board cells:
  (for* ([row (in-range 25)]
         [col (in-range 25)])
    (send dc set-origin (* col SIZE) (* row SIZE))
    (let ([c (board-ref row col)])
      (cond
        [(eq? c 'out)
         (send dc set-brush "black" 'solid)]
        [(room? c)
         (send dc set-brush "lightblue" 'solid)]
        [else
         (send dc set-brush "lightgray" 'solid)])
      (send dc draw-rectangle 0 0 (sub1 SIZE) (sub1 SIZE))
      (when (room? c)
        (center-string "white" (symbol->string c)))
      (when (door? c)
        (send dc set-brush "lightblue" 'solid)
        (let ([n (board-ref (- row 1) col)]
              [s (board-ref (+ row 1) col)]
              [e (board-ref row (+ col 1))]
              [w (board-ref row (- col 1))]
              [r (door->room c)])
          (cond
            [(eq? r n)
             (send dc draw-rectangle 0 0 (sub1 SIZE) (quotient SIZE 2))]
            [(eq? r s)
             (send dc draw-rectangle 0 (/ SIZE 2) (sub1 SIZE) (quotient SIZE 2))]
            [(eq? r e)
             (send dc draw-rectangle (/ SIZE 2) 0 (- (sub1 SIZE) (quotient SIZE 2)) (sub1 SIZE))]
            [(eq? r w)
             (send dc draw-rectangle 0 0 (- (sub1 SIZE) (quotient SIZE 2)) (sub1 SIZE))])))))
  ;; Draw player dots:
  (for ([pos (in-list (state->pos-list current-state))]
        [dest (in-list current-state)]
        [i (in-naturals)]
        [color (in-list '("red" "yellow" "white" "green" "blue" "purple"))])
    (send dc set-origin (* (cadr pos) SIZE) (* (car pos) SIZE))
    (send dc set-brush color 'solid)
    (send dc set-pen "BLACK" 1 (if (eq? (cadr dest) 'S) 'solid 'transparent))
    (send dc draw-ellipse 2 2 (- SIZE 5) (- SIZE 5))
    (center-string "black" (number->string i)))
  ;; Offscreen to screen:
  (send canvas-dc draw-bitmap (send dc get-bitmap) 0 0))
      
;; Need a lock, since board state is shared between
;;  GUI thread and server thread:
(define lock-sema (make-semaphore 1))
(define advance-sema (make-semaphore))
(define current-state null)

;; Create the GUI window:
(define f (new frame%
               [label "Clue"]))
(define pane (new horizontal-pane% 
                  [parent f]
                  [alignment '(center center)]))
(define next
  (new button% 
       [label "Next" ]
       [parent pane]
       [enabled #f]
       [callback (lambda (b e)
                   (semaphore-post advance-sema)
                   (send next enable #f))]))
(define auto
  (new check-box% 
       [label "Auto next"]
       [parent pane]
       [callback (lambda (c e)
                   (when (send auto get-value)
                     (send next enable #f)
                     (semaphore-post advance-sema)))]))
(define c (new canvas% 
               [parent f]
               [style '(no-autoclear)]
               [paint-callback
                (lambda (c canvas-dc)
                  (call-with-semaphore
                   lock-sema
                   (lambda ()
                     (draw-board canvas-dc)
                     (if (send auto get-value)
                         (semaphore-post advance-sema)
                         (send next enable #t)))))]))
(send c min-client-width (* 25 SIZE))
(send c min-client-height (* 25 SIZE))
(define msg (new message%
                 [label " "]
                 [parent f]
                 [stretchable-width #t]))
(define status (new message%
                    [label " "]
                    [parent f]
                    [stretchable-width #t]))
(send f show #t)

(define msg-accum "")

;; Start the server and players:
(void
 (play-all (lambda (str)
             (set! msg-accum
                   (if (string=? msg-accum "")
                         str
                         (string-append msg-accum "; " str)))
             (send msg set-label msg-accum))
           (lambda (str state)
             (send status set-label str)
             (set! msg-accum "")
             (call-with-semaphore
              lock-sema
              (lambda ()
                (set! advance-sema (make-semaphore))
                (set! current-state state)
                (send c refresh)))
             (semaphore-wait advance-sema))))

;; Wait as long as the GUI window is open:
(void (yield 'wait))

;; Terminate all threads:
(custodian-shutdown-all (current-custodian))
